from django.conf import settings

GCM_DEVICE_MODEL = getattr(settings, 'GCM_DEVICE_MODEL', 'gcm.models.Device')

GCM_APIKEY = getattr(settings, 'GCM_APIKEY', None)

GCM_MAX_RECIPIENTS = getattr(settings, 'GCM_MAX_RECIPIENTS', 1000)



from __future__ import unicode_literals
import json
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

from django.contrib.auth import get_user_model
from django.core import management
from django.core.exceptions import ImproperlyConfigured
from django.core.management import CommandError
from django.core.urlresolvers import reverse
from django.test import TestCase
from tastypie.test import ResourceTestCase
from mock import patch, PropertyMock, MagicMock
import requests

from . import conf
from .models import get_device_model
from .api import GCMMessage as ApiGCMMessage

Device = get_device_model()
User = get_user_model()


class CommandTest(TestCase):

    def test_gcm_urls(self):
        out = StringIO()
        management.call_command('gcm_urls', stdout=out)
        self.assertIn('/gcm/v1/device/register', out.getvalue())
        self.assertIn('/gcm/v1/device/unregister', out.getvalue())

    @patch.object(ApiGCMMessage, 'send')
    def test_send_message(self, mock_send):
        device_name = 'My test device'
        device = Device.objects.create(dev_id='device_1', name=device_name,
                                       reg_id='000123abc001', is_active=True)

        mock_send.return_value = (
            Device.objects.values_list('reg_id', flat=True),
            {'failure': 0, 'canonical_ids': 0, 'success': 1,
             'multicast_id': 112233, 'results': [{'message_id': '0:123123'}]})

        out = StringIO()
        management.call_command('gcm_messenger', device.id, 'test', stdout=out)
        self.assertTrue(mock_send.called)

        management.call_command('gcm_messenger', devices=True, stdout=out)
        self.assertIn(device_name, out.getvalue())

        self.assertRaises(
            CommandError, management.call_command, 'gcm_messenger')
        self.assertRaises(
            CommandError, management.call_command,
            'gcm_messenger', '999', 'test')


class AdminTest(TestCase):

    def setUp(self):
        user_password = 'password'
        user = User.objects.create_superuser(
            'admin', 'admin@test.com', user_password)
        self.client.login(username=user.username, password=user_password)

    @patch.object(ApiGCMMessage, 'send')
    def test_send_message(self, mock_send):
        device = Device.objects.create(
            dev_id='device_1', reg_id='000123abc001', is_active=True)

        mock_send.return_value = (
            Device.objects.values_list('reg_id', flat=True),
            {'failure': 0, 'canonical_ids': 0, 'success': 1,
             'multicast_id': 112233, 'results': [{'message_id': '0:123123'}]})

        self.client.post('/admin/gcm/device/',
                         data={'action': 'send_message_action',
                               '_selected_action': device.id})

        response = self.client.post('/admin/gcm/device/send-message/',
                                    data={'message': 'admin test message'})

        self.assertTrue(mock_send.called)
        self.assertEqual(response.status_code, 302)

    def test_do_not_send_empty_message(self):
        device = Device.objects.create(
            dev_id='device_1', reg_id='000123abc001', is_active=True)

        self.client.post('/admin/gcm/device/',
                         data={'action': 'send_message_action',
                               '_selected_action': device.id})

        response = self.client.post('/admin/gcm/device/send-message/')
        self.assertEqual(response.status_code, 200)

    def test_send_message_view_requires_devices(self):
        response = self.client.get('/admin/gcm/device/send-message/')
        self.assertRedirects(response, '/admin/gcm/device/')


class DeviceResourceTest(ResourceTestCase):

    def setUp(self):
        super(DeviceResourceTest, self).setUp()
        url_kwargs = {'resource_name': 'device', 'api_name': 'v1'}
        self.api_register_url = reverse(
            "register-device", kwargs=url_kwargs)
        self.api_unregister_url = reverse(
            "unregister-device", kwargs=url_kwargs)

    def _not_allowed_methods(self, url):
        self.assertHttpMethodNotAllowed(self.api_client.get(url))
        self.assertHttpMethodNotAllowed(self.api_client.put(url))
        self.assertHttpMethodNotAllowed(self.api_client.delete(url))

    def test_register_device_not_allowed_methods(self):
        self._not_allowed_methods(self.api_register_url)

    def test_unregister_device_not_allowed_methods(self):
        self._not_allowed_methods(self.api_unregister_url)

    def test_register_device(self):
        device_id = 'TEST001'
        data = {
            'dev_id': device_id,
            'reg_id': 'abcd1234',
        }
        response = self.api_client.post(self.api_register_url, data=data)
        self.assertHttpOK(response)

        devices = Device.objects.filter(dev_id=device_id)
        self.assertEqual(devices.count(), 1)
        self.assertTrue(list(devices).pop().is_active)

    def test_unregister_device(self):
        device_id = 'TEST001'
        data = {
            'dev_id': device_id,
        }
        response = self.api_client.post(self.api_unregister_url, data=data)
        self.assertHttpOK(response)

        devices = Device.objects.filter(dev_id=device_id)
        self.assertEqual(devices.count(), 1)
        self.assertFalse(list(devices).pop().is_active)

    def test_register_device_without_id(self):
        response = self.api_client.post(self.api_register_url, data={})
        self.assertHttpBadRequest(response)

    def test_update_registration_id(self):
        device_id = 'TEST'
        expected_registration_id = 'xyz1'
        device = Device.objects.create(dev_id=device_id, reg_id='abc1')

        data = {
            'dev_id': device_id,
            'reg_id': expected_registration_id
        }
        response = self.api_client.post(self.api_register_url, data=data)
        self.assertHttpOK(response)

        reg_id = Device.objects.get(pk=device.pk).reg_id
        self.assertEqual(reg_id, expected_registration_id)


class GCMMessageTest(TestCase):

    @patch.object(ApiGCMMessage, 'send')
    def test_mark_inactive(self, mock_send):
        Device.objects.create(
            dev_id='device_1', reg_id='000123abc001', is_active=True)
        Device.objects.create(
            dev_id='device_2', reg_id='000123abc002', is_active=True)
        Device.objects.create(
            dev_id='device_3', reg_id='000123abc003', is_active=True)

        mock_send.return_value = (
            Device.objects.values_list('reg_id', flat=True),
            {'failure': 2, 'canonical_ids': 0, 'success': 1,
             'multicast_id': 112233, 'results': [
                 {'error': 'InvalidRegistration'},
                 {'message_id': '0:123123'},
                 {'error': 'NotRegistered'}]})

        Device.objects.all().send_message('test message')

        devices = Device.objects.filter(is_active=False)
        self.assertEqual(devices.count(), 2)

    @patch.object(ApiGCMMessage, 'send')
    def test_ignore_unhandled_error(self, mock_send):
        Device.objects.create(
            dev_id='device_1', reg_id='000123abc001', is_active=True)

        mock_send.return_value = (
            Device.objects.values_list('reg_id', flat=True),
            {'failure': 1, 'canonical_ids': 0, 'success': 0,
             'multicast_id': 112233, 'results': [{'error': 'UnhandledError'}]})

        Device.objects.all().send_message('test message')

        device = Device.objects.get(dev_id='device_1')
        self.assertTrue(device.is_active)

    @patch.object(ApiGCMMessage, 'send')
    def test_ignore_active_device(self, mock_send):
        dev_id = 'device_1'
        device = Device.objects.create(
            dev_id=dev_id, reg_id='000123abc001', is_active=True)

        mock_send.return_value = (
            Device.objects.values_list('reg_id', flat=True),
            {'failure': 0, 'canonical_ids': 0, 'success': 1,
             'multicast_id': 112233, 'results': [{'message_id': '0:123123'}]})

        device.send_message('test message')
        self.assertEqual(str(Device.objects.get(is_active=True)), dev_id)

    @patch.object(ApiGCMMessage, 'send')
    def test_ignore_empty_queryset(self, mock_send):
        Device.objects.all().send_message('test')
        self.assertFalse(mock_send.called)

    @patch.object(conf, 'GCM_MAX_RECIPIENTS',
                  new_callable=PropertyMock(return_value=2))
    def test_split_to_chunks(self, mock_max_recipients):

        Device.objects.create(
            dev_id='device_1', reg_id='000123abc001', is_active=True)
        Device.objects.create(
            dev_id='device_2', reg_id='000123abc002', is_active=True)

        Device.objects.create(
            dev_id='device_3', reg_id='000123abc003', is_active=True)
        Device.objects.create(
            dev_id='device_4', reg_id='000123abc004', is_active=True)

        Device.objects.create(
            dev_id='device_5', reg_id='000123abc005', is_active=True)

        chunk_messages = [
            {'failure': 1, 'canonical_ids': 0, 'success': 0,
             'multicast_id': '0003', 'results': [{'error': 'NotRegistered'}]},
            {'failure': 1, 'canonical_ids': 0, 'success': 1,
             'multicast_id': '0002', 'results': [
                 {'message_id': '0:123123'},
                 {'error': 'InvalidRegistration'}]},
            {'failure': 1, 'canonical_ids': 0, 'success': 1,
             'multicast_id': '0001', 'results': [
                 {'error': 'InvalidRegistration'},
                 {'message_id': '0:123123'}]}]

        def side_effect(**kwargs):
            mock = MagicMock()
            mock.content = json.dumps(chunk_messages.pop())
            return mock

        with patch.object(requests, 'post', side_effect=side_effect):
            Device.objects.all().send_message('test message')

        devices = Device.objects.filter(is_active=False)
        self.assertEqual(devices.count(), 3)

    @patch.object(conf, 'GCM_APIKEY',
                  new_callable=PropertyMock(return_value=None))
    def test_configuration(self, mock_apikey):
        device = Device.objects.create(
            dev_id='device_1', reg_id='000123abc001', is_active=True)
        self.assertRaises(
            ImproperlyConfigured, device.send_message, data='test')



import requests
import json

from django.core.exceptions import ImproperlyConfigured

from . import conf


class GCMMessage(object):

    def __init__(self):
        self.api_key = conf.GCM_APIKEY

        if not self.api_key:
            raise ImproperlyConfigured(
                "You haven't set the 'GCM_APIKEY' setting yet.")

    def _chunks(self, items, limit):
        """
        Yield successive chunks from list \a items with a minimum size \a limit
        """
        for i in range(0, len(items), limit):
            yield items[i:i + limit]

    def send(self, regs_id, data, **kwargs):
        """
        Send a GCM message for one or more devices, using json data
        regs_id: A list with the devices which will be receiving a message
        data: The dict data which will be send
        Optional params e.g.:
            collapse_key: A string to group messages
        For more info see the following documentation:
        https://developer.android.com/google/gcm/server-ref.html#send-downstream
        """

        if not isinstance(data, dict):
            data = {'msg': data}

        if len(regs_id) > conf.GCM_MAX_RECIPIENTS:
            ret = []
            for chunk in self._chunks(regs_id, conf.GCM_MAX_RECIPIENTS):
                ret.append(self.send(chunk, data, **kwargs))
            return ret

        values = {
            'registration_ids': regs_id,
            'data': data,
            'collapse_key': 'message'}
        values.update(kwargs)

        values = json.dumps(values)

        headers = {
            'UserAgent': "GCM-Server",
            'Content-Type': 'application/json',
            'Authorization': 'key=' + self.api_key}

        response = requests.post(url="https://android.googleapis.com/gcm/send",
                                 data=values,
                                 headers=headers)
        response.raise_for_status()
        return regs_id, json.loads(response.content)



from django.core.management.base import NoArgsCommand
from django.core.urlresolvers import reverse


class Command(NoArgsCommand):
    help = "Show GCM urls"

    def show_line(self):
        self.stdout.write("%s\n" % ("-" * 30))

    def handle_noargs(self, **options):
        url_kwargs = {'resource_name': 'device', 'api_name': 'v1'}
        register_url = reverse("register-device", kwargs=url_kwargs)
        unregister_url = reverse("unregister-device", kwargs=url_kwargs)

        self.show_line()
        self.stdout.write("GCM urls:\n")
        self.stdout.write("* Register device\n    %s\n" % register_url)
        self.stdout.write("* Unregister device\n    %s\n" % unregister_url)
        self.show_line()



from django.db import models
from django.db.models.query import QuerySet
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from . import conf
from . import api
from .utils import load_object


def get_device_model():
    return load_object(conf.GCM_DEVICE_MODEL)


class GCMMessage(api.GCMMessage):
    GCM_INVALID_ID_ERRORS = ['InvalidRegistration',
                             'NotRegistered',
                             'MismatchSenderId']

    def send(self, regs_id, data, **kwargs):
        response = super(GCMMessage, self).send(regs_id, data, **kwargs)
        chunks = [response] if not isinstance(response, list) else response
        for chunk in chunks:
            self.post_send(*chunk)
        return response

    def post_send(self, regs_id, response):
        if response['failure']:
            invalid_messages = dict(filter(
                lambda x: x[1].get('error') in self.GCM_INVALID_ID_ERRORS,
                zip(regs_id, response.get('results'))))

            regs = list(invalid_messages.keys())
            for device in get_device_model().objects.filter(reg_id__in=regs):
                device.mark_inactive(
                    error_message=invalid_messages[device.reg_id]['error'])


class DeviceQuerySet(QuerySet):

    def send_message(self, data, **kwargs):
        if self:
            return GCMMessage().send(
                regs_id=list(self.values_list("reg_id", flat=True)),
                data=data, **kwargs)


class DeviceManager(models.Manager):

    def get_queryset(self):
        return DeviceQuerySet(self.model)
    get_query_set = get_queryset  # Django < 1.6 compatiblity


@python_2_unicode_compatible
class AbstractDevice(models.Model):

    dev_id = models.CharField(
        verbose_name=_("Device ID"), max_length=50, unique=True,)
    reg_id = models.CharField(
        verbose_name=_("Registration ID"), max_length=255, unique=True)
    name = models.CharField(
        verbose_name=_("Name"), max_length=255, blank=True, null=True)
    creation_date = models.DateTimeField(
        verbose_name=_("Creation date"), auto_now_add=True)
    modified_date = models.DateTimeField(
        verbose_name=_("Modified date"), auto_now=True)
    is_active = models.BooleanField(
        verbose_name=_("Is active?"), default=False)

    objects = DeviceManager()

    def __str__(self):
        return self.dev_id

    class Meta:
        abstract = True
        verbose_name = _("Device")
        verbose_name_plural = _("Devices")
        ordering = ['-modified_date']

    def send_message(self, data, **kwargs):
        return GCMMessage().send(regs_id=[self.reg_id], data=data, **kwargs)

    def mark_inactive(self, **kwargs):
        self.is_active = False
        self.save()


class Device(AbstractDevice):
    pass



from django.conf.urls import url
from django.core.exceptions import ObjectDoesNotExist
from django.http import HttpResponse, HttpResponseBadRequest
from tastypie.authentication import Authentication
from tastypie.authorization import Authorization
from tastypie.resources import Resource
from tastypie.serializers import Serializer

from . import signals
from .forms import RegisterDeviceForm, UnregisterDeviceForm
from .models import get_device_model


class DeviceResource(Resource):

    DEVICE_ID_FIELD_NAME = 'dev_id'

    model_class = get_device_model()
    register_form_class = RegisterDeviceForm
    unregister_form_class = UnregisterDeviceForm
    object = None
    request = None

    class Meta:
        resource_name = 'device'
        allowed_methods = ['post']
        authentication = Authentication()
        authorization = Authorization()
        serializer = Serializer(formats=['json'])

    def prepend_urls(self):
        return [
            url(r"^(?P<resource_name>%s)/register/$" %
                self._meta.resource_name, self.wrap_view('register'),
                name="register-device"),
            url(r"^(?P<resource_name>%s)/unregister/$" %
                self._meta.resource_name, self.wrap_view('unregister'),
                name="unregister-device"),
        ]

    def _verify(self, request):
        self.is_authenticated(request)
        self.throttle_check(request)
        self.method_check(request, self._meta.allowed_methods)

    def get_form_kwargs(self):
        data = self.deserialize(self.request, self.request.body)
        kwargs = {'data': data}
        try:
            kwargs['instance'] = self.get_instance(data)
        except ObjectDoesNotExist:
            pass
        return kwargs

    def get_form(self, form_class):
        return form_class(**self.get_form_kwargs())

    def get_queryset(self):
        return self.model_class.objects.all()

    def get_instance(self, data):
        kwargs = {self.DEVICE_ID_FIELD_NAME:
                  data.get(self.DEVICE_ID_FIELD_NAME)}
        return self.get_queryset().get(**kwargs)

    def form_valid(self, form):
        self.object = form.save()
        return HttpResponse

    def form_invalid(self, form):
        return HttpResponseBadRequest

    def _form_processing(self, form_class, request):
        self._verify(request)
        self.request = request

        form = self.get_form(form_class)

        if form.is_valid():
            response_class = self.form_valid(form)
        else:
            response_class = self.form_invalid(form)

        return self.create_response(request, data={},
                                    response_class=response_class)

    def register(self, request, **kwargs):
        form_class = self.register_form_class
        response = self._form_processing(form_class, request)
        signals.device_registered.send(sender=self, device=self.object,
                                       request=request)
        return response

    def unregister(self, request, **kwargs):
        form_class = self.unregister_form_class
        response = self._form_processing(form_class, request)
        signals.device_unregistered.send(sender=self, device=self.object,
                                         request=request)
        return response



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from south.utils import datetime_utils as datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):

        # Changing field 'Device.reg_id'
        db.alter_column(u'gcm_device', 'reg_id', self.gf('django.db.models.fields.CharField')(unique=True, max_length=255))

    def backwards(self, orm):
        pass

    models = {
        u'gcm.device': {
            'Meta': {'ordering': "['-modified_date']", 'object_name': 'Device'},
            'creation_date': ('django.db.models.fields.DateTimeField', [], {'auto_now_add': 'True', 'blank': 'True'}),
            'dev_id': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '50'}),
            u'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'is_active': ('django.db.models.fields.BooleanField', [], {'default': 'False'}),
            'modified_date': ('django.db.models.fields.DateTimeField', [], {'auto_now': 'True', 'blank': 'True'}),
            'name': ('django.db.models.fields.CharField', [], {'max_length': '255', 'null': 'True', 'blank': 'True'}),
            'reg_id': ('django.db.models.fields.CharField', [], {'unique': 'True', 'max_length': '255'})
        }
    }

    complete_apps = ['gcm']


from django.conf.urls import patterns, url, include
from tastypie.api import Api

from .resources import DeviceResource

gcm_api = Api(api_name='v1')
gcm_api.register(DeviceResource())


urlpatterns = patterns('', url(r'^gcm/', include(gcm_api.urls)))



# ~*~ coding: utf-8 ~*~
from setuptools import setup, find_packages
from gcm import VERSION

setup(
    name='django-gcm',
    version=VERSION,
    description='Google Cloud Messaging Server',
    long_description=open('README.rst').read(),
    author='Adam Bogdał',
    author_email='adam@bogdal.pl',
    url='https://github.com/bogdal/django-gcm',
    packages=find_packages(),
    package_data={
        'gcm': ['locale/*/LC_MESSAGES/*']
    },
    include_package_data=True,
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules'],
    zip_safe=False,
    install_requires=[
        'django>=1.6',
        'django-tastypie>=0.10.0',
        'pytz>=2013.8',
        'requests>=1.2.0',
    ],
    extras_require={
        'dev': [
            'mock>=1.0.1',
            'coverage'
        ]
    }
)



from django.utils.importlib import import_module


def load_object(object_path):

    module_path, object_name = object_path.rsplit('.', 1)
    module = import_module(module_path)

    return getattr(module, object_name)
